from typing import Union

import uvicorn
from fastapi import FastAPI

from steroid.constants import APP_COMPONENT, APP_COMPONENT_TYPE, APP_CONTROLLER
from steroid.controller import Controller
from steroid.logging import getLogConfig, getLogger
from steroid.utils import clearTerminal


class App:
    _APP: FastAPI = None
    logger = None

    def __new__(cls, *args, **kwargs):
        if cls._APP is None:
            cls.logger = getLogger()
            cls._APP = FastAPI(*args, **kwargs)

        return cls

    @classmethod
    @property
    def app(cls):
        return cls._APP

    @classmethod
    def start(cls, host: str = "0.0.0.0", port: int = 8000):
        clearTerminal()

        uvicorn.run(cls.app, host=host, port=port, log_config=getLogConfig())

    @classmethod
    def addController(cls, controller):
        componentType = getattr(controller, APP_COMPONENT_TYPE, None)
        component: Union[Controller, None] = getattr(controller, APP_COMPONENT, None)

        if not componentType == APP_CONTROLLER:
            raise Exception(f"Controller component must be of type {APP_CONTROLLER}")

        if not component:
            raise Exception(f"Controller component not detected")

        component.setupController(cls.app)



APP_COMPONENT = "APP_COMPONENT"
APP_COMPONENT_TYPE = "APP_COMPONENT_TYPE"

APP_CONTROLLER = "APP_CONTROLLER"
APP_CONTROLLER_METHOD = "APP_CONTROLLER_METHOD"
APP_METHOD_MIDDLEWARE = "APP_METHOD_MIDDLEWARE"
APP_MIDDLEWARE_TARGET = "APP_MIDDLEWARE_TARGET"



import inspect

from steroid.constants import APP_COMPONENT, APP_COMPONENT_TYPE, APP_CONTROLLER_METHOD
from steroid.utils import formatPath, removeLeadingOrTrailingSlash


class BaseMethod:
    _METHOD: str = None
    _PATH: str = None
    _FULL_PATH: str = None

    _CONTROLLER = None

    def __init__(self, path: str = "", *args, **kwargs):
        self._PATH = formatPath(path)

        self.args = args
        self.kwargs = kwargs

    def __call__(self, func):
        self.func = func

        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        setattr(wrapper, APP_COMPONENT_TYPE, APP_CONTROLLER_METHOD)
        setattr(wrapper, APP_COMPONENT, self)

        return wrapper

    def setupMethod(self, controller):
        self._CONTROLLER = controller

        httpMethodDecorator = getattr(controller.router, self._METHOD.lower())
        parameters = [
            parameter
            for parameter in dict(inspect.signature(self.func).parameters).values()
        ]

        @httpMethodDecorator(self.path)
        def method():
            return self.func(*parameters)

        return method

    @property
    def path(self):
        if len(self._PATH) > 0:
            return f"/{removeLeadingOrTrailingSlash(self._PATH)}/"
        return "/"


class Get(BaseMethod):
    _METHOD = "GET"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Post(BaseMethod):
    _METHOD = "POST"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Put(BaseMethod):
    _METHOD = "PUT"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Patch(BaseMethod):
    _METHOD = "PATCH"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Delete(BaseMethod):
    _METHOD = "DELETE"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import logging
from logging.config import dictConfig

from pydantic import BaseModel


class LogConfig(BaseModel):
    """Logging configuration to be set for the server"""

    DATE_FORMAT: str = "%d/%m/%Y, %H:%M:%S"

    LOG_NAME: str = "steroid"
    LOG_FORMAT: str = "[%(asctime)s] %(levelname)s %(message)s"
    LOG_LEVEL: str = "INFO"

    UVICORN_ACCESS_LOG_FORMAT: str = '[%(asctime)s] %(levelname)s [Uvicorn] %(client_addr)s - "%(request_line)s" %(status_code)s'

    version: int = 1
    disable_existing_loggers: bool = False
    formatters: dict = {
        "default": {
            "()": "uvicorn.logging.DefaultFormatter",
            "fmt": LOG_FORMAT,
            "datefmt": DATE_FORMAT,
        },
        "access": {
            "()": "uvicorn.logging.AccessFormatter",
            "fmt": UVICORN_ACCESS_LOG_FORMAT,
            "datefmt": DATE_FORMAT,
            "use_colors": False,
        },
    }
    handlers: dict = {
        "default": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "access": {
            "formatter": "access",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "error": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
    }
    loggers: dict = {
        LOG_NAME: {
            "handlers": ["default"],
            "level": LOG_LEVEL,
        },
        "uvicorn.access": {
            "handlers": ["access"],
            "level": LOG_LEVEL,
        },
        "uvicorn.error": {
            "handlers": ["error"],
            "level": LOG_LEVEL,
        },
    }


def getLogConfig():
    return LogConfig().model_dump()


logger = None


def getLogger(disableAccessLog: bool = False, disableErrorLog: bool = True):
    """
    If the logger is not created yet, create it.
    Otherwise, return the logger.
    This is to avoid creating multiple loggers.
    This is a singleton pattern.
    https://en.wikipedia.org/wiki/Singleton_pattern
    """

    global logger

    if not logger:
        dictConfig(getLogConfig())
        logger = logging.getLogger(LogConfig().LOG_NAME)

        logging.getLogger("uvicorn.access").disabled = disableAccessLog
        logging.getLogger("uvicorn.error").disabled = disableErrorLog

    return logger



import inspect
from typing import Literal, Union

from fastapi import HTTPException, Request
from fastapi.responses import JSONResponse

from steroid.constants import (
    APP_COMPONENT,
    APP_COMPONENT_TYPE,
    APP_METHOD_MIDDLEWARE,
    APP_MIDDLEWARE_TARGET,
)
from steroid.utils import getApp, removeLeadingOrTrailingSlash


class Middleware:
    HOOK: Union[Literal["PRE"], Literal["POST"]] = "PRE"

    def __call__(self, cls):
        return cls

    def action(request: Request):
        raise NotImplementedError("Middleware action must be implemented")


class UseMiddlewares:
    _APPLY_TO_PATH: str = None
    _APPLY_TO_METHOD: str = None

    _PRE_MIDDLEWARES: list[Middleware] = []
    _POST_MIDDLEWARES: list[Middleware] = []

    def __init__(self, *middlewares: list[Middleware]):
        self._PRE_MIDDLEWARES = [
            middleware for middleware in middlewares if middleware.HOOK == "PRE"
        ]
        self._POST_MIDDLEWARES = [
            middleware for middleware in middlewares if middleware.HOOK == "POST"
        ]

    def __call__(self, func):
        if not inspect.isfunction(func):
            raise Exception("Middlewares can only be applied to functions")

        self.func = func

        setattr(self.func, APP_MIDDLEWARE_TARGET, getattr(func, APP_COMPONENT, None))
        setattr(self.func, APP_COMPONENT_TYPE, APP_METHOD_MIDDLEWARE)
        setattr(self.func, APP_COMPONENT, self)

        return self.func

    def setupMiddleware(self):
        method = getattr(self.func, APP_MIDDLEWARE_TARGET, None)

        if method:
            self._APPLY_TO_PATH = f"{removeLeadingOrTrailingSlash(method._CONTROLLER.path)}/{removeLeadingOrTrailingSlash(method._PATH)}"
            self._APPLY_TO_METHOD = method._METHOD

        async def _handleMiddlewareRoutes(request: Request, call_next):
            try:
                self.handleMiddlewares(request, self._PRE_MIDDLEWARES)

                response = await call_next(request)

                self.handleMiddlewares(request, self._POST_MIDDLEWARES)

                return response
            except HTTPException as error:
                return JSONResponse(
                    status_code=error.status_code, content={"detail": error.detail}
                )

        getApp().middleware("http")(_handleMiddlewareRoutes)

        return self.func

    def handleMiddlewares(self, request: Request, middlewares: list[Middleware]):
        path = removeLeadingOrTrailingSlash(request.url.path)
        method = request.method

        if method == self._APPLY_TO_METHOD and path == self._APPLY_TO_PATH:
            for middleware in middlewares:
                middleware.action(request)



import inspect

from fastapi.routing import APIRouter

from steroid.constants import (
    APP_COMPONENT,
    APP_COMPONENT_TYPE,
    APP_CONTROLLER,
    APP_CONTROLLER_METHOD,
    APP_METHOD_MIDDLEWARE,
    APP_MIDDLEWARE_TARGET,
)
from steroid.logging import getLogger
from steroid.utils import formatPath, removeLeadingOrTrailingSlash

logger = getLogger()


class Controller:
    def __init__(self, path: str = "", *args, **kwargs):
        self.router = APIRouter()

        self.path = formatPath(path)

        self.args = args
        self.kwargs = kwargs

    def __call__(self, cls):
        self.cls = cls

        self.mapComponentsToController(cls)

        setattr(cls, APP_COMPONENT_TYPE, APP_CONTROLLER)
        setattr(cls, APP_COMPONENT, self)

        return cls(*self.args, **self.kwargs)

    def setupController(self, app):
        app.include_router(
            self.router,
            prefix=self.path,
            tags=[removeLeadingOrTrailingSlash(self.path)] if self.path else [],
        )

    def mapComponentsToController(self, classObject):
        logger.info("[Controller] %s {%s}" % (self.cls.__name__, self.path))

        for _, method in classObject.__dict__.items():
            if inspect.isfunction(method):
                componentType = getattr(method, APP_COMPONENT_TYPE, None)
                component = getattr(method, APP_COMPONENT, None)
                if componentType == APP_CONTROLLER_METHOD:
                    self.handleAppControllerMethod(component)
                elif componentType == APP_METHOD_MIDDLEWARE:
                    self.handleAppMethodMiddleware(component)

    def handleAppControllerMethod(self, component):
        setupControllerMethod = getattr(component, "setupMethod", None)
        if setupControllerMethod:
            setupControllerMethod(self)

            logger.info(
                "[Method] Mapped {%s%s, %s} route"
                % (self.path, component._PATH, component._METHOD)
            )

    def handleAppMethodMiddleware(self, component):
        middlewareTarget = getattr(component.func, APP_MIDDLEWARE_TARGET, None)
        self.handleAppControllerMethod(middlewareTarget)

        setupMiddleware = getattr(component, "setupMiddleware", None)
        if setupMiddleware:
            setupMiddleware()



import os
from subprocess import call


def formatPath(text: str) -> str:
    if len(text) == 0:
        return text
    return f"/{text}" if not text.startswith("/") else text


def removeLeadingOrTrailingSlash(text: str) -> str:
    if text.startswith("/"):
        text = text[1:]
    if text.endswith("/"):
        text = text[:-1]

    return text


def getApp():
    from steroid.app import App

    return App().app


def clearTerminal():
    print("\033c", end="")
